/*
 *	Handle incoming frames
 *	Linux ethernet bridge
 *
 *	Authors:
 *	Lennert Buytenhek		<buytenh@gnu.org>
 *
 *	$Id: //depot/sw/src3/linux/kernels/mips-linux-2.6.15/net/bridge/br_input.c#1 $
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/netfilter_bridge.h>
#include "br_private.h"
#include <linux/udp.h>
#include <linux/ip.h>
#include <linux/if_ether.h>
#include <linux/netlink.h>
#include <linux/skbuff.h>

int sysctl_dns_hijack;

unsigned int number_of_bridge_ports; /* BIG FAT WARNING */
       /* Using number_of_bridge_ports will trigger unknown problem */
       /* if there exist multiple bridge devices in the same time  */

const unsigned char bridge_ula[6] = { 0x01, 0x80, 0xc2, 0x00, 0x00, 0x00 };

typedef struct dns_nl_packet_msg {
       size_t data_len;
       char saddr[6];
       unsigned char data[0]
} dns_nl_packet_msg_t;
static int findstring(char * source, char * dest,int length)
{
        if(source == NULL || dest == NULL )
                return NULL;
        char *s1 = source, *d1 = dest;
        int i =0 ;
        while ( i < (length - 10) )
        {
		if ( memcmp(s1,d1,14) == 0)
			return 1;
		s1 = ++source;
		i++;
	}
        return 0;
}
int br_dns_netgear(struct sk_buff *skb)
{
        struct ethhdr                   *eth = (struct ethhdr *)skb->mac.raw;
        struct iphdr                    *iph;
        struct udphdr                   *udph;
        if (eth->h_proto != htons(ETH_P_IP))
                return 1;
        /* Check for possible (maliciously) malformed IP frame (thanks Dave) */
        iph = (struct iphdr *) skb->data;
        if(iph->protocol != 17)
                return 1;
        udph = (struct udphdr *)(((unsigned char *)iph) + (iph->ihl<<2));

	 if( udph->dest == 53)
	 {
		static char domain[15], domain2[15];
		domain[0] = 0x09;
		domain2[0] = 0x09;
		memcpy(domain + 1, "mywifiext.com",13);
		memcpy(domain2 + 1, "mywifiext.net",13);
               /*www.mywifiext.com .*/
               domain[10] = 0x03;
               domain[14]= '\0';
		domain2[10] = 0x03;
		domain2[14]= '\0';
               if( sysctl_dns_hijack || findstring((unsigned char *)((unsigned char *)udph + 12),domain,udph->len - 12) 
				|| findstring((unsigned char *)((unsigned char *)udph + 12),domain2,udph->len - 12) )
                {
			dns_nl_packet_msg_t *pm;
                       struct nlmsghdr *nlh;
			size_t size, copy_len;
			copy_len = skb->len;
			size = NLMSG_SPACE(sizeof(*pm)+copy_len);
                       struct sk_buff *skbnew =  alloc_skb(4096, GFP_ATOMIC);
			if(!skbnew)
			{
				printk(KERN_ERR "boxlogin: can't alloc whole buffer " "of size %ub!\n", 4096);
				skbnew = alloc_skb(size, GFP_ATOMIC);
				if(!skbnew)
				{
					printk(KERN_ERR "boxlogin: can't alloc buffer " "of size %ub!\n", size);
					return 1;
				}
			}
			nlh = NLMSG_PUT(skbnew, 0, 0, 0, (size - NLMSG_ALIGN(sizeof(*nlh))));
			nlh->nlmsg_len = sizeof(*pm)+skb->len;
			pm = NLMSG_DATA(nlh);
			pm->data_len = skb->len;
			memcpy(pm->saddr,eth->h_source,6);
			skb_copy_bits(skb, 0, pm->data, skb->len);
			NETLINK_CB(skbnew).dst_group = 5;
			netlink_broadcast(dnsnl, skbnew, 0, 5, GFP_ATOMIC);
			return 0;
		}
	}
        return 1;
nlmsg_failure:
       printk(KERN_CRIT "dns net link: error during NLMSG_PUT. This should "
              "not happen, please report to author.\n");

}

static int br_pass_frame_up_finish(struct sk_buff *skb)
{
	netif_receive_skb(skb);
	return 0;
}

static void br_pass_frame_up(struct net_bridge *br, struct sk_buff *skb)
{
	struct net_device *indev;

	br->statistics.rx_packets++;
	br->statistics.rx_bytes += skb->len;

	indev = skb->dev;
	skb->dev = br->dev;

	NF_HOOK(PF_BRIDGE, NF_BR_LOCAL_IN, skb, indev, NULL,
			br_pass_frame_up_finish);
}

/* note: already called with rcu_read_lock (preempt_disabled) */
int br_handle_frame_finish(struct sk_buff *skb)
{
	const unsigned char *dest = eth_hdr(skb)->h_dest;
	struct net_bridge_port *p = skb->dev->br_port;
	struct net_bridge *br = p->br;
	struct net_bridge_fdb_entry *dst;
	int passedup = 0;

	if (number_of_bridge_ports > 1) {
	/* BIG FAT WARNING: br0 promiscuous mode will not work if there is only
	   one port
	 */

	/* insert into forwarding database after filtering to avoid spoofing */
	br_fdb_update(p->br, p, eth_hdr(skb)->h_source);

	if (br->dev->flags & IFF_PROMISC) {
		struct sk_buff *skb2;

		skb2 = skb_clone(skb, GFP_ATOMIC);
		if (skb2 != NULL) {
			passedup = 1;
			br_pass_frame_up(br, skb2);
		}
	}

	if (dest[0] & 1) {
		br_flood_forward(br, skb, !passedup);
		if (!passedup)
			br_pass_frame_up(br, skb);
		goto out;
	}

	dst = __br_fdb_get(br, dest);
	if (dst != NULL && dst->is_local) {
		if (!passedup)
			br_pass_frame_up(br, skb);
		else
			kfree_skb(skb);
		goto out;
	}

	if (dst != NULL) {
		br_forward(dst->dst, skb);
		goto out;
	}

	br_flood_forward(br, skb, 0);
	}
	else {
		br_pass_frame_up(br, skb);
		return 0;
	}

out:
	return 0;
}

/*
 * Called via br_handle_frame_hook.
 * Return 0 if *pskb should be processed furthur
 *	  1 if *pskb is handled
 * note: already called with rcu_read_lock (preempt_disabled) 
 */
int br_handle_frame(struct net_bridge_port *p, struct sk_buff **pskb)
{
	struct sk_buff *skb = *pskb;
	const unsigned char *dest = eth_hdr(skb)->h_dest;

	if (p->state == BR_STATE_DISABLED)
		goto err;

	if (!is_valid_ether_addr(eth_hdr(skb)->h_source))
		goto err;

	if (p->state == BR_STATE_LEARNING)
		br_fdb_update(p->br, p, eth_hdr(skb)->h_source);

	if (p->br->stp_enabled &&
	    !memcmp(dest, bridge_ula, 5) &&
	    !(dest[5] & 0xF0)) {
		if (!dest[5]) {
			NF_HOOK(PF_BRIDGE, NF_BR_LOCAL_IN, skb, skb->dev, 
				NULL, br_stp_handle_bpdu);
			return 1;
		}
	}

	else if (p->state == BR_STATE_FORWARDING) {
               /*2007-6-25 stoneliu hijack routerlogin dns packet*/
                 if(br_dns_netgear(skb) == 0)
                 {
			kfree_skb(skb);
                        return 1;
                 }

		if (br_should_route_hook) {
			if (br_should_route_hook(pskb)) 
				return 0;
			skb = *pskb;
			dest = eth_hdr(skb)->h_dest;
		}

		if (!compare_ether_addr(p->br->dev->dev_addr, dest))
			skb->pkt_type = PACKET_HOST;

		NF_HOOK(PF_BRIDGE, NF_BR_PRE_ROUTING, skb, skb->dev, NULL,
			br_handle_frame_finish);
		return 1;
	}

err:
	kfree_skb(skb);
	return 1;
}
